/* ----------------------------------------------------------------------- *
 *
 *   Copyright 2009 Koji Suzuki - All Rights Reserved
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, Inc., 53 Temple Place Ste 330,
 *   Boston MA 02111-1307, USA; either version 2 of the License, or
 *   (at your option) any later version; incorporated herein by reference.
 *
 * ----------------------------------------------------------------------- */

#include <windows.h>
#include <ddk/ntddstor.h>
#include <getopt.h>
#include <stdio.h>
#include <ctype.h>

#define SECTOR_SIZE		512
#define MAX_MBR_SECS		32

struct partition_record {
	unsigned char bootable;
	unsigned char first_sector[3];
	unsigned char fs_desc;
	unsigned char last_sector[3];
	unsigned char first_sector_numbers[4];
	unsigned char number_of_sectors[4];
};

static inline unsigned int load3(unsigned char *p) {
	return p[0] + (p[1]<<8) + (p[2]<<16);
}

static inline unsigned int load4(unsigned char *p) {
	return p[0] + (p[1]<<8) + (p[2]<<16) + (p[3]<<24);
}

int cmp_mode = 0; // 0: write mbr, 1: compare mode
int mbr_size;
char drive_letter;

char *mbr_path = "grldr.mbr";
unsigned char dk_buf[SECTOR_SIZE * MAX_MBR_SECS];
unsigned char mbr_buf[SECTOR_SIZE * MAX_MBR_SECS];

void win32_error(char *);

void usage() {
	printf("usage:\n");
	printf("  grldrmbr [-c] [-f mbr_path] drive_letter:\n");
}

int xcmp(unsigned char *a, unsigned char *b, int len) {
	int i;
	for (i=0; i<len; i++) {
		if (*a++ != *b++) return i;
	}
	return -1;
}

int main(int argc,char *argv[]) {
	int c;
	HANDLE handle;
	BOOL r;
	STORAGE_DEVICE_NUMBER sdn;
	DWORD count;
	DWORD rcount;
	struct partition_record *pr;
	char path[128];

	while ((c = getopt(argc, argv, "cf:")) != EOF) {
		switch(c) {
		case 'c':
			cmp_mode = 1;
			break;
		case 'f':
			mbr_path = optarg;
			break;
		default:
			usage();
			exit(1);
		}
	}
	if (optind != argc - 1) {
		usage();
		exit(1);
	}
	drive_letter = argv[optind][0];
	if (argv[optind][1] != ':' || argv[optind][2] != 0 || 
	      (!('a' <= drive_letter && drive_letter <= 'z')
	    && !('A' <= drive_letter && drive_letter <= 'Z'))) {
		usage();
		exit(1);
	}
	sprintf(path, "\\\\.\\%c:",drive_letter);
	handle = CreateFile(path, GENERIC_READ, FILE_SHARE_READ
			,NULL,OPEN_EXISTING,0,NULL);	
	if ( handle == INVALID_HANDLE_VALUE ) {
		win32_error("Accessing logical drive");
		exit(1);
	}
	if (!DeviceIoControl( handle, IOCTL_STORAGE_GET_DEVICE_NUMBER, NULL,
		0, (LPVOID)(&sdn), sizeof(sdn), &rcount, NULL)
			|| sizeof(sdn) != rcount) {
		win32_error("DeviceIoControl/GET_DEVICE_NUMBER failed");
		exit(1);
	}
	CloseHandle(handle);
	printf("logical drive   %c:\n",drive_letter);
	printf("physical drive  %d (type %d)\n"
			,sdn.DeviceNumber, sdn.DeviceType);

	sprintf(path, "\\\\.\\PHYSICALDRIVE%d",sdn.DeviceNumber);

	handle = CreateFile(path, GENERIC_READ , FILE_SHARE_READ 
			,NULL,OPEN_EXISTING,0,NULL);

	if ( handle == INVALID_HANDLE_VALUE ) {
		win32_error("Accessing physical drive");
		exit(1);
	}
	count = SECTOR_SIZE * MAX_MBR_SECS;
	if (!ReadFile(handle, dk_buf, count, &rcount, NULL)
			|| count != rcount ) {
		win32_error("Reading physical drive");
		exit(1);
	}
	CloseHandle(handle);

	handle = CreateFile(mbr_path, GENERIC_READ , FILE_SHARE_READ 
			,NULL,OPEN_EXISTING,0,NULL);
	if ( handle == INVALID_HANDLE_VALUE ) {
		win32_error("open mbr file");
		exit(1);
	}
	count = SECTOR_SIZE * MAX_MBR_SECS;
	if (!ReadFile(handle, mbr_buf, count, &rcount, NULL)) {
		win32_error("Reading mbr file");
		exit(1);
	}
	CloseHandle(handle);
	mbr_size = rcount;
	printf("mbr file size : %d (%d sectors)  file name %s\n"
			,mbr_size,(mbr_size+SECTOR_SIZE-1)/SECTOR_SIZE
			, mbr_path);
	if (dk_buf[0x1fe] != 0x55 || dk_buf[0x1ff] != 0xaa) {
		printf("disk %d: has no patition table -- format first\n");
		exit(1);
	}
	if (mbr_buf[0x1fe] != 0x55 || mbr_buf[0x1ff] != 0xaa) {
		printf("%s is not mbr file\n",mbr_path);
		exit(1);
	}
	pr = (struct partition_record *)(dk_buf+446);
	printf("first sector  : %d\n",load4(pr->first_sector_numbers));
	if ((mbr_size+SECTOR_SIZE-1)/SECTOR_SIZE >= MAX_MBR_SECS) {
		printf("mbr file is too large (up to %d sectors)\n",MAX_MBR_SECS - 1);
		exit(1);
	}
	if ((mbr_size+SECTOR_SIZE-1)/SECTOR_SIZE >= load4(pr->first_sector_numbers)) {
		printf("mbr file is too large (over first sector)\n");
		exit(1);
	}
	if (cmp_mode) {
		if (memcmp(dk_buf, mbr_buf, 0x1b8)) {
			printf("compare error at offset %d\n",xcmp(dk_buf, mbr_buf, 0x1b8));
			exit(1);
		}
		if (memcmp(dk_buf + 0x200, mbr_buf + 0x200, mbr_size - 0x200)) {
			printf("compare error at offset %d\n",xcmp(dk_buf + 0x200, mbr_buf + 0x200, mbr_size - 0x200)+0x200);
			exit(1);
		}
		printf("compare OK\n");
		exit(0);
	}
	// yesno("really ?");
	memcpy(dk_buf, mbr_buf, 0x1b8);
	memcpy(dk_buf + 0x200, mbr_buf + 0x200, mbr_size - 0x200);
	handle = CreateFile(path, GENERIC_READ | GENERIC_WRITE
			, FILE_SHARE_READ | FILE_SHARE_WRITE
			,NULL,OPEN_EXISTING,FILE_FLAG_NO_BUFFERING,NULL);

	if ( handle == INVALID_HANDLE_VALUE ) {
		win32_error("Accessing physical drive");
		exit(1);
	}
	if (!WriteFile(handle, dk_buf, mbr_size, &rcount, NULL)
			|| mbr_size != rcount ) {
		win32_error("Writing physical drive");
		exit(1);
	}
	CloseHandle(handle);
	printf("OK\n");
	exit(0);
}

void win32_error(char* msg) {
	LPVOID lpMsgBuf;

	FormatMessage( 
                FORMAT_MESSAGE_ALLOCATE_BUFFER |
                FORMAT_MESSAGE_FROM_SYSTEM |
                FORMAT_MESSAGE_IGNORE_INSERTS,
                NULL, GetLastError(),
                MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
                (LPTSTR) &lpMsgBuf, 0, NULL );
	fprintf(stderr, "%s: %s", msg, (char*) lpMsgBuf);
	LocalFree(lpMsgBuf);
}

/* memo :(from grun4dos/README)

How to write GRLDR.MBR to the Master Boot Track of a hard disk?

First, read the Windows disk signature and partition information bytes
(72 bytes in total, from offset 0x01b8 to 0x01ff of the MBR sector), and put
them on the same range from offset 0x01b8 to 0x01ff of the beginning sector of
GRLDR.MBR.

*/
